<?php
/**
 * @copyright: Copyright © 2020 Firebear Studio. All rights reserved.
 * @author   : Firebear Studio <fbeardev@gmail.com>
 */

namespace Firebear\PlatformSap\Model\Export\Adapter\Gateway;

use Firebear\ImportExport\Api\Export\History\CompressInterface;
use Firebear\ImportExport\Logger\Logger;
use Firebear\ImportExport\Model\Filesystem\File\ReadFactory;
use Firebear\ImportExport\Model\Source\Factory as SourceFactory;
use Firebear\ImportExport\Traits\General as GeneralTrait;
use Firebear\PlatformSap\Model\LogRepository;
use Magento\Customer\Model\ResourceModel\AddressRepository;
use Magento\Customer\Model\ResourceModel\CustomerRepository;
use Magento\Framework\App\CacheInterface;
use Magento\Framework\App\Config\ScopeConfigInterface;
use Magento\Framework\Exception\LocalizedException;
use Magento\Framework\Exception\NoSuchEntityException;
use Magento\Framework\Filesystem;
use Magento\Framework\Filesystem\Directory\WriteFactory as DirectoryWriteFactory;
use Magento\Framework\Filesystem\File\WriteFactory as FileWriteFactory;
use Magento\Framework\ObjectManagerInterface;
use Magento\Framework\Stdlib\DateTime\Timezone;
use Magento\Sales\Model\OrderRepository;
use Symfony\Component\Console\Output\ConsoleOutput;
use Magento\Directory\Model\CountryFactory;
use Magento\Sales\Model\Order\Status\HistoryFactory;
use Firebear\PlatformSap\Model\Import\Source\Gateway\OrderItems as OrderItemsGateway;

/**
 * Sap order gateway
 */
class Order extends SalesAbstractGateway
{
    use GeneralTrait;

    const SAP_ENTITY_ENDPOINT = 'SalesOrder/Create';

    /**
     * @var OrderRepository
     */
    protected $orderRepository;

    /**
     * @var CustomerRepository
     */
    protected $customerRepository;

    /**
     * @var AddressRepository
     */
    protected $addressRepository;

    /**
     * @var ObjectManagerInterface
     */
    protected $objectManager;

    /**
     * @var Customer
     */
    protected $customerGateway;

    /**
     * @var Invoice
     */
    protected $invoiceGateway;

    /**
     * @var Creditmemo
     */
    protected $creditmemoGateway;

    /**
     * @var HistoryFactory
     */
    protected $statusHistoryFactory;

    /**
     * @var OrderItemsGateway
     */
    protected $orderItemsGateway;

    /**
     * Order constructor.
     * @param ScopeConfigInterface $scopeConfig
     * @param OrderRepository $orderRepository
     * @param Logger $logger
     * @param ConsoleOutput $output
     * @param Filesystem $filesystem
     * @param ReadFactory $readFactory
     * @param DirectoryWriteFactory $writeFactory
     * @param FileWriteFactory $fileWrite
     * @param Timezone $timezone
     * @param SourceFactory $factory
     * @param CacheInterface $cache
     * @param LogRepository $logRepository
     * @param CustomerRepository $customerRepository
     * @param AddressRepository $addressRepository
     * @param ObjectManagerInterface $objectManager
     * @param CompressInterface $compressCommand
     * @param Customer $customerGateway
     * @param CountryFactory $countryFactory
     * @param HistoryFactory $statusHistoryFactory
     * @param Invoice $invoiceGateway
     * @param Creditmemo $creditmemoGateway
     * @param OrderItemsGateway $orderItemsGateway
     */
    public function __construct(
        ScopeConfigInterface $scopeConfig,
        OrderRepository $orderRepository,
        Logger $logger,
        ConsoleOutput $output,
        Filesystem $filesystem,
        ReadFactory $readFactory,
        DirectoryWriteFactory $writeFactory,
        FileWriteFactory $fileWrite,
        Timezone $timezone,
        SourceFactory $factory,
        CacheInterface $cache,
        LogRepository  $logRepository,
        CustomerRepository $customerRepository,
        AddressRepository $addressRepository,
        ObjectManagerInterface $objectManager,
        CompressInterface $compressCommand,
        \Firebear\PlatformSap\Model\Export\Adapter\Gateway\Customer $customerGateway,
        CountryFactory $countryFactory,
        HistoryFactory $statusHistoryFactory,
        Invoice $invoiceGateway,
        Creditmemo $creditmemoGateway,
        OrderItemsGateway $orderItemsGateway
    ) {
        parent::__construct(
            $scopeConfig,
            $filesystem,
            $readFactory,
            $writeFactory,
            $fileWrite,
            $timezone,
            $factory,
            $cache,
            $logRepository,
            $compressCommand,
            $countryFactory
        );
        $this->orderRepository = $orderRepository;
        $this->customerRepository = $customerRepository;
        $this->addressRepository = $addressRepository;
        $this->_logger = $logger;
        if (empty($this->_logger->getFileName())) {
            $this->_logger->setFileName('sapOrderUpdate');
        }
        $this->output = $output;
        $this->objectManager = $objectManager;
        $this->customerGateway = $customerGateway;
        $this->statusHistoryFactory = $statusHistoryFactory;
        $this->invoiceGateway = $invoiceGateway;
        $this->creditmemoGateway = $creditmemoGateway;
        $this->orderItemsGateway = $orderItemsGateway;
    }

    /**
     * @param $orderId
     * @param array $data
     * @throws LocalizedException
     * @throws NoSuchEntityException
     * @throws \Magento\Framework\Exception\AlreadyExistsException
     * @throws \Magento\Framework\Exception\InputException
     */
    public function addOrder($orderId, $data = [])
    {
        $orderRepository = $this->objectManager->create(\Magento\Sales\Model\OrderRepository::class);
        $customerRegistry = $this->objectManager->create(\Magento\Customer\Model\CustomerRegistry::class);
        $addressRegistry = $this->objectManager->create(\Magento\Customer\Model\AddressRegistry::class);
        $order = $orderRepository->get($orderId);
        $taxCode = $this->scopeConfig->getValue('firebear_importexport/sap/sap_default_tax_code') ?? 'Exempt';
        $customerId = $order->getCustomerId();
        $sapId = (int) $order->getSapId() > 0 ? (int) $order->getSapId() : '';
        $sapEntityData = [
            'entity' => 'order',
            'entity_id' => $order->getEntityId(),
            'sap_id' => $sapId,
            'customer_id' => $customerId,
            'customer_name' => $order->getCustomerFirstname() . ' ' . $order->getCustomerLastname(),
            'email' => $order->getCustomerEmail(),
            'increment_id' => $order->getIncrementId(),
            'delivery_date' => date('Y-m-d', strtotime('+4 days')),
            'shipping_amount' => $order->getShippingAmount(),
            'ship_to' => '',
            'bill_to' => '',
            'status' => $order->getStatus(),
            'tax_code' => $taxCode
        ];

        $payment = $order->getPayment();
        if ($payment) {
            $sapEntityData['payment'] = $payment->getMethod();
            $sapEntityData['po_number'] = $payment->getPoNumber();
        }

        if ($customerId) {
            $customer = $customerRegistry->retrieve($customerId);

            $sapId = $customer->getSapId();
            if (!empty($sapId)) {
                $sapEntityData['customer_sap_id'] = $sapId;
            }
        }

        if ($order->getShippingAddress()->getCustomerAddressId()) {
            $address = false;
            $addressSapId = false;
            try {
                $address = $addressRegistry->retrieve($order->getShippingAddress()->getCustomerAddressId())->getDataModel();
                $addressSapId = $address->getCustomAttribute('sap_id');
            } catch (\Exception $e) {
            }

            if (!empty($addressSapId)) {
                $sapEntityData['ship_to'] = $addressSapId->getValue();
            } elseif ($address) {
                $sapEntityData['ship_to'] = 'Ship To #' . $order->getShippingAddress()->getCustomerAddressId();
            } else {
                $sapEntityData['ship_to'] = '';
            }
        }
        if ($order->getBillingAddress()->getCustomerAddressId()) {
            $address = false;
            $addressSapId = false;
            try {
                $address = $addressRegistry->retrieve($order->getBillingAddress()->getCustomerAddressId())->getDataModel();
                $addressSapId = $address->getCustomAttribute('sap_id');
            } catch (\Exception $e) {
            }

            if (!empty($addressSapId)) {
                $sapEntityData['bill_to'] = $addressSapId->getValue();
            } elseif ($address) {
                $sapEntityData['bill_to'] = 'Bill To #' . $order->getBillingAddress()->getCustomerAddressId();
            } else {
                $sapEntityData['bill_to'] = '';
            }
        }

        $lineNum = 0;
        foreach ($order->getItems() as $item) {
            $sapItem = [
                'line_num' => $lineNum++,
                'sku' => $item->getSku(),
                'product_name' => $item->getName(),
                'qty' => floatval($item->getQtyOrdered()),
                'price' => floatval($item->getPrice()),
                'discount_percent' => floatval($item->getDiscountPercent()),
                'tax_code' => $taxCode,
            ];
            if ($item->getDiscountPercent() == 0 && $item->getDiscountAmount() > 0) {
                $discountPercent = $item->getDiscountAmount() * 100 / $item->getPrice();
                $sapItem['discount_percent'] = $discountPercent;
            }
            $sapEntityData['items'][] = $sapItem;
        }

        if ($order->getCustomerId()) {
            $customerData = [
                'entity' => 'customer',
                'email' => $order->getCustomerEmail()
            ];
            try {
                $result = $this->customerGateway->exportCustomer($customerData);
            } catch (\Exception $e) {
            }
        }

        //$sapEntityData['customer'] = $this->getCustomerData($order);
        $sapEntityData['addresses'] = $this->getOrderAddresses($order);
        $sapEntityData['shipping_address_text'] = $sapEntityData['addresses']['shipping_address_text'];
        $sapEntityData['billing_address_text'] = $sapEntityData['addresses']['billing_address_text'];
        unset($sapEntityData['addresses']['shipping_address_text']);
        unset($sapEntityData['addresses']['billing_address_text']);

        if ($this->scopeConfig->getValue('firebear_importexport/sap/sap_export_shipment')) {
            $sapEntityData['shipment'] = $this->getOrderShipment($order);
        }

        $sapEntityData['status_history'] = $this->getOrderStatusHistory($order);

        $sapEntityData = $this->prepareOrderData($sapEntityData);

        $updateResponse = $this->uploadSourceData($sapEntityData);

        if (!$updateResponse['error']) {
            if (isset($updateResponse['object_id'])) {
                $order->setData('sap_id', $updateResponse['object_id']);
                $order->setData('sap_status', 'ok');
                $statusHistory = $this->statusHistoryFactory->create();
                $statusHistory->setStatus($order->getStatus());

                $message = '';
                foreach ($sapEntityData['items'] as $sapEntityItem) {
                    $message = $message . $sapEntityItem['sku'] .
                        '&emsp;'. $sapEntityItem['product_name'].
                        '&emsp;'. $sapEntityItem['price'].
                        '&emsp;(' . $sapEntityItem['qty'] .')<br><b>sync to SBO OK.</b>';
                }

                $statusHistory->setComment($message);
                $order->addStatusHistory($statusHistory);
                $this->orderRepository->save($order);
            }

            $this->addLogWriteln($updateResponse['message'], $this->output);

            if ($this->scopeConfig->getValue('firebear_importexport/sap/sap_export_invoice')) {
                $this->exportInvoices($order);
            }

            if ($this->scopeConfig->getValue('firebear_importexport/sap/sap_export_creditmemo')) {
                $this->exportCreditMemo($order);
            }

        } else {
            $statusHistory = $this->statusHistoryFactory->create();
            $statusHistory->setStatus($order->getStatus());
            $statusHistory->setComment($updateResponse['message']);
            $order->addStatusHistory($statusHistory);
            $order->setData('sap_status', 'failed');
            $this->orderRepository->save($order);
            $this->addLogWriteln($updateResponse['message'], $this->output, 'error');
        }
    }

    /**
     * @param $order
     * @return array
     */
    protected function getOrderStatusHistory($order)
    {
        $result = [];
        $statusHistories = $order->getStatusHistories();
        if ($statusHistories) {
            foreach ($statusHistories as $exportStatusHistory) {
                $statusHistoryData = $exportStatusHistory->getData();
                $statusHistory = [
                    'status' => $statusHistoryData['status'],
                    'comment' => $statusHistoryData['comment'],
                    'created_at' => $statusHistoryData['created_at']
                ];
                $result[] = $statusHistory;
            }
        }
        return $result;
    }

    /**
     * @param $order
     * @return array
     */
    protected function exportInvoices($order)
    {
        $result = [];
        $invoices = $order->getInvoiceCollection()->getItems();
        if ($invoices) {
            foreach ($invoices as $exportInvoice) {
                try {
                    $sapId = (int) $exportInvoice->getSapId() > 0 ? (int) $exportInvoice->getSapId() : '';
                    if (empty($sapId)) {
                        $this->invoiceGateway->addInoivce($order,$exportInvoice);
                    }
                } catch (\Exception $e) {
                }
            }
        }
        return $result;
    }

    /**
     * @param $order
     * @return array
     */
    protected function exportCreditMemo($order)
    {
        $result = [];
        $creditMemos = $order->getCreditmemosCollection()->getItems();
        if ($creditMemos) {
            foreach ($creditMemos as $creditMemo) {
                try {
                    $sapId = (int) $creditMemo->getSapId() > 0 ? (int) $creditMemo->getSapId() : '';
                    if (empty($sapId)) {
                        $this->creditmemoGateway->addCreditMemo($order,$creditMemo);
                    }
                } catch (\Exception $e) {
                }
            }
        }
        return $result;
    }

    /**
     * @param $order
     * @return array
     */
    protected function getOrderShipment($order)
    {
        $result = [];
        $shipments = $order->getShipmentsCollection();
        if ($shipments) {
            foreach ($shipments as $exportShipments) {
                $shipmentData = $exportShipments->getData();
                $shipmentItems = [];

                foreach ($exportShipments->getAllItems() as $shipmentItem) {
                    $item = [
                        'weight' => $shipmentItem->getWeight(),
                        'qty' => floatval($shipmentItem->getQty()),
                        'sku' => $shipmentItem->getSku()
                    ];
                    $shipmentItems[] = $item;
                }

                $shipment = [
                    'entity' => $shipmentData['entity_id'],
                    'total_qty' => floatval($shipmentData['total_qty']),
                    'total_weight' => floatval($shipmentData['total_weight']),
                    'created_at' => $shipmentData['created_at'],
                    'updated_at' => $shipmentData['updated_at'],
                    'items' => $shipmentItems
                ];
                $result[] = $shipment;
            }
        }
        return $result;
    }

    /**
     *  The function to modify the order data in the custom extensions
     *
     * @param $data
     * @return mixed
     */
    public function prepareOrderData($data)
    {
        return $data;
    }

    /**
     * @param \Magento\Sales\Api\Data\OrderInterface $order
     * @return array
     * @throws \Magento\Framework\Exception\LocalizedException
     * @throws \Magento\Framework\Exception\NoSuchEntityException
     */
    protected function getCustomerData(\Magento\Sales\Api\Data\OrderInterface $order)
    {
        $customerData = [];
        try {
            $customerId = $order->getCustomerId();
            if (!empty($customerId)) {
                try {
                    $customer = $this->customerRepository->getById($customerId);
                    $customerData = $this->getCustomerDataFromEntity($customer);
                } catch (NoSuchEntityException $e) {
                    $customerData = $this->getCustomerDataFromOrder($order);
                }
            } else {
                $customerData = $this->getCustomerDataFromOrder($order);
            }
        } catch (LocalizedException $e) {
            return $customerData;
        }
        return $customerData;
    }

    /**
     * @param \Magento\Customer\Api\Data\CustomerInterface $customer
     * @return array
     */
    protected function getCustomerDataFromEntity(\Magento\Customer\Api\Data\CustomerInterface $customer)
    {
        $customerData = [];
        $customerData['entity'] = 'customer';
        $customerData['entity_id'] = $customer->getId();
        $customerData['firstname'] = $customer->getFirstname();
        $customerData['lastname'] = $customer->getLastname();
        $customerData['email'] = $customer->getEmail();
        $customerData['taxvat'] = $customer->getTaxvat();

        $sapId = $customer->getCustomAttribute('sap_id');
        if (!empty($sapId)) {
            $sapId = $sapId->getValue();
        }
        $customerData['sap_id'] = $sapId;

        foreach ($customer->getAddresses() as $customerAddress) {
            $billingAddress = $this->prepareAddress($customerAddress, 'billing');
            $shippingAddress = $this->prepareAddress($customerAddress, 'shipping');

            if (!empty($billingAddress['telephone'])) {
                $customerData['telephone'] = $billingAddress['telephone'];
            }
            $customerData['addresses'][] = $billingAddress;
            $customerData['addresses'][] = $shippingAddress;
        }
        return $customerData;
    }

    /**
     * @param \Magento\Customer\Api\Data\AddressInterface $customerAddress
     * @param $type
     * @return mixed
     */
    protected function prepareAddress(\Magento\Customer\Api\Data\AddressInterface $customerAddress, $type)
    {
        $sapId = $customerAddress->getCustomAttribute('sap_id');
        if (!empty($sapId)) {
            $sapId = $sapId->getValue();
        }
        $address['entity_id'] = $customerAddress->getId();
        $address['sap_id'] = $sapId;
        $address['postcode'] = $customerAddress->getPostcode();
        $address['city'] = $customerAddress->getCity();
        $street = $customerAddress->getStreet();
        $address['street_0'] = isset($street[0]) ? $street[0] : '';
        $address['street_1'] = isset($street[1]) ? $street[1] : '';
        $address['country_id'] = $customerAddress->getCountryId();
        $address['region_id'] = substr($customerAddress->getRegion()->getRegionCode(), 0, 2);
        $address['vat_id'] = $customerAddress->getVatId();
        $address['telephone'] = $customerAddress->getTelephone();
        $address['type'] = $type;
        $address['default'] = 0;

        if ($type == 'billing' && $customerAddress->isDefaultBilling()) {
            $address['default'] = 1;
        }
        if ($type == 'shipping' && $customerAddress->isDefaultShipping()) {
            $address['default'] = 1;
        }

        return $address;
    }

    /**
     * @param \Magento\Sales\Api\Data\OrderInterface $order
     * @return array
     */
    protected function getCustomerDataFromOrder(\Magento\Sales\Api\Data\OrderInterface $order)
    {
        $customerData = [];
        $customerData['entity'] = 'customer';
        $customerData['firstname'] = $order->getCustomerFirstname();
        $customerData['lastname'] = $order->getCustomerLastname();
        $customerData['email'] = $order->getCustomerEmail();

        $billingAddress = $order->getBillingAddress();
        if ($billingAddress) {
            $customerData['addresses'][] = $this->prepareOrderAddress($billingAddress, 'billing');
        }

        $shippingAddress = $order->getShippingAddress();
        if ($shippingAddress) {
            $customerData['addresses'][] = $this->prepareOrderAddress($shippingAddress, 'shipping');
        }
        return $customerData;
    }

    /**
     * @param \Magento\Sales\Api\Data\OrderAddressInterface $orderAddress
     * @param $type
     * @return array
     */
    protected function prepareOrderAddress(\Magento\Sales\Api\Data\OrderAddressInterface $orderAddress, $type)
    {
        $street = $orderAddress->getStreet();
        $address = [
            'postcode' => $orderAddress->getPostcode(),
            'city' => $orderAddress->getCity(),
            'street_0' => isset($street[0]) ? $street[0] : '',
            'street_1' => isset($street[1]) ? $street[1] : '',
            'country_id' => $orderAddress->getCountryId(),
            'region_id' => substr($orderAddress->getRegionCode(), 0, 2),
            'vat_id' => $orderAddress->getVatId(),
            'type' => 'billing',
            'default' => 1
        ];

        if ($type == 'billing') {
            $address['type'] = 'billing';
        }
        if ($type == 'shipping') {
            $address['type'] = 'shipping';
        }
        return $address;
    }
}
